//===-- temu-c/Assembler.h - Assembler and Dissassebler----------*- C++ -*-===//
//
// T-EMU: The Terma Emulator
// (c) Terma 2015
// Authors: Mattias Holm <maho (at) terma.com>
//
//===----------------------------------------------------------------------===//
///
/// \file
///
/// \brief This file contains the declarations of the assembler and
/// disassembler.
///
//===----------------------------------------------------------------------===//

#ifndef TEMU_SUPPORT_ASSEMBLER
#define TEMU_SUPPORT_ASSEMBLER
#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include "temu-c/Support/Attributes.h"
/*! Assemble an instruction
 * \param Cpu processor for which to assemble.
 * \param AsmStr String with instruction in the CPUs assembler.
 * \result Instruction bitpattern.
 */
TEMU_API uint32_t temu_assemble(void *Cpu, const char *AsmStr);

/*! Disassemble an instruction.
 * \warning The function returns a malloced string. You have ownership
 *           and is responsible for calling free() on it.
 * \result A pointer to a malloced instruction string
 *         or null incase of failure.
 */
TEMU_API char* temu_disassemble(void *Cpu, uint32_t Instr);

/*! Disassemble an instruction.
 * \warning The function returns a thread local string. It is borrowed
 * and not allowed to escape the current thread without a copy.
 * \result A pointer to a a thread local instruction string
 *         or null incase of failure.
 */
TEMU_API const char* temu_disassembleAuto(void *Cpu, uint32_t Instr);

/*! Assemble an instruction to memory
 * \param Cpu processor for which to assemble.
 * \param AsmStr String with instruction in the CPUs assembler.
 * \param Addr Physical address of where to put the instruction.
 */
TEMU_API void temu_assembleToMemory(void *Cpu, const char *AsmStr, uint64_t Addr);

/*! Disassemble an instruction in memory
 *
 * The function dissasmbles an instruction and retuns a string allocated
 * on the heap. You are responsible for its release using free().
 *
 * \warning The function returns a malloced string. You have ownership
 *           and are responsible for calling free() on it.
 *
 * \param Cpu CPU whose memory space will be used for disasembling.
 * \param Addr Physical address of instruction to dissassemble.
 * \result A pointer to a malloced instruction string
 *         or null incase of failure.
 */
TEMU_API char* temu_disassembleMemory(void *Cpu, uint64_t Addr);

// Returns a thread local string instead
TEMU_API const char* temu_disassembleMemoryAuto(void *Cpu, uint64_t Addr);

#ifdef __cplusplus
}
#endif

#endif /* ! TEMU_SUPPORT_ASSEMBLER */
