#ifndef TEMU_REGISTER_H
#define TEMU_REGISTER_H
#include <stdint.h>
#include "temu-c/Support/Attributes.h"
#include "temu-c/Support/Objsys.h"
#include "temu-c/Memory/Memory.h"

/*
 * EXPERIMENTAL API
 * API Changes:
 *   2018-06-11: Remove wr and rd arguments from temu_addField
 *               Replace the separate flag params with a flag bit field.
 */

// Registers are normal properties suitable for device registers (must
// be of an integer type), but have addtitional meta-data attached to
// them. Primarily you can associate fields with a register, a
// register field may even have individual accessors which can be
// invoked to set single bits.

#ifdef __cplusplus
extern "C" {
#endif

typedef struct temu_RegisterBank temu_RegisterBank;
typedef struct temu_Register temu_Register;
typedef struct temu_Field temu_Field;

/*!
 * \brief temu_addRegisterBank Adds a register bank to a T-EMU class
 * \param C Pointer to the T-EMU class
 * \param Name Name of the register
 * \param MemAccessIface Pointer to the register memory access interface
 * \result Pointer to the new register bank
 */
TEMU_API temu_RegisterBank *
temu_addRegisterBank(temu_Class *C, const char *Name,
                     temu_MemAccessIface *MemAccessIface);

/*!
 * \brief temu_getRegisterBankName Retrieves the register bank name from its
 * object \param Bank Pointer to the register bank name \result Bank name as a
 * C-string
 */
TEMU_API const char *temu_getRegisterBankName(temu_RegisterBank *Bank);

/*!
 * Add register property to class
 *
 * Adds a register with the given name to a class. It returns a reference
 * to a meta register which can be used to add fields.
 *
 * \param Bank Register bank to add register to
 * \param Name Name of register (must be a valid C-identifier)
 * \param Offset Offset to storage element in the device struct.
 * \param Typ Type of register, note that registers are limited to
 *        unsigned integer types with fixed width.
 * \param Count Number of registers (normally 1)
 * \param Wr Register write function
 * \param Rd Register read function
 * \param Doc Documentation string for register.
 *
 * \param DeviceOffset Offset of register in memory system. This is
 *        the same offset that is used in the memory transaction
 *        interface.
 *
 * \param Stride In case the register is an array of registers, then
 *        the stride is used for the offset in physical address space
 *        between each register.
 * \result A reference to a meta register which can be used to add fields.
 */
TEMU_API temu_Register *
temu_addRegister(temu_RegisterBank *Bank, const char *Name, int Offset,
                 temu_Type Typ, int Count, temu_PropWriter Wr,
                 temu_PropReader Rd, const char *Doc,
                 // Offset of register from device mapping (or bank mapping)
                 // in memory space
                 uint32_t DeviceOffset, uint32_t Stride);

#define TEMU_FIELD_WR (1 << 0)
#define TEMU_FIELD_WARM_RESET (1 << 1)

/*!
 * Add field to meta register
 *
 * Adds a field to the meta register.
 *
 * \param R Meta register pointer.
 * \param Name Name of of field (must be a C-compatible identifier)
 * \param Mask Mask identifying the bits in the register corresponding
 *        to the field. Mask must contain consecutive bits only.
 * \param ColdReset Reset value used for a cold reset.
 * \param WarmReset Reset value used for a warm reset.
 * \param Flags Flags used: TEMU_FIELD_WR means field is writable,
 *        TEMU_FIELD_WARM_RESET means the field is subject to reset actions
 *        also on warm resets.
 * \param Doc Documentation string
 */

TEMU_API void temu_addField(temu_Register *R, const char *Name, uint64_t Mask,
                            uint64_t ColdReset, uint64_t WarmReset,
                            uint64_t Flags, const char *Doc);

/*!
 * \brief temu_getRegisterBank Get a register bank from a class by name
 * \param C Pointer to the T-EMU class
 * \param Name Name of the register bank
 * \result Pointer to the register bank
 */
TEMU_API temu_RegisterBank *temu_getRegisterBank(temu_Class *C,
                                                 const char *Name);

/*!
 * \brief temu_getRegisterBanks Get all register banks of a class
 * \param C Pointer to the T-EMU class
 * \param Idx The index, at which the reading of banks should start
 * \param Length The length to be read
 * \param Banks Array of pointers to the register banks, must be allocated with
 * size "Length" \result zero on success, otherwise non-zero value
 */
TEMU_API int temu_getRegisterBanks(temu_Class *C, unsigned Idx, unsigned Length,
                                   temu_RegisterBank **Banks);

/*!
 * \brief temu_getRegister Get a register from a register bank by name
 * \param Bank Pointer to the register bank object
 * \param Name Name of the register
 * \result Pointer to the register object
 */
TEMU_API temu_Register *temu_getRegister(temu_RegisterBank *Bank,
                                         const char *Name);

/*!
 * \brief temu_getRegisters Get all or chunk of registers from a bank
 * \param Bank Pointer to the register bank object
 * \param Idx The starting number of the register
 * \param Length The length of the
 * \param Regs Array of pointers to the registers to be allocated with size
 * "Length" \result number of registers read
 */
TEMU_API int temu_getRegisters(temu_RegisterBank *Bank, unsigned Idx,
                               unsigned Length, temu_Register **Regs);

/*!
 * \brief temu_getRegisterName Retrieve register name from pointer
 * \param R Pointer to the register
 * \result The name of the register
 */
TEMU_API const char *temu_getRegisterName(temu_Register *R);

/*!
 * \brief temu_getRegisterDocs Get the documentation of a register as a string
 * \param R Pointer to the register
 * \result The documentation of the register as a C-string
 */
TEMU_API const char *temu_getRegisterDocs(temu_Register *R);

/*!
 * \brief temu_getField Get a field of a register
 * \param R Pointer to the register that contains the field
 * \param Name Name of the field in question
 * \return Pointer to the field object
 */
TEMU_API temu_Field *temu_getField(temu_Register *R, const char *Name);

/*!
 * \brief temu_getFields Get all fields of a register
 * \param R Pointer to the register that has the fields
 * \param Idx The starting index at which a field will be read in the full array
 * \param Length The length to be read
 * \param Fields Array of pointers to the fields objects, must be allocated with
 * size "Length" \result Number of fields read
 */
TEMU_API int temu_getFields(temu_Register *R, unsigned Idx, unsigned Length,
                            temu_Field **Fields);

/*!
 * \brief temu_getFieldName Get the name of a field from a field object
 * \param R Pointer to the field object
 * \return Name of the field as a C-string
 */
TEMU_API const char *temu_getFieldName(temu_Field *R);

/*!
 * \brief temu_getFieldDocs Get the documentation of a field as a C-string
 * \param R Pointer to the field object
 * \return The documentation of the field as a C-string
 */
TEMU_API const char *temu_getFieldDocs(temu_Field *R);

/*!
 * \brief temu_getFieldMask Get field mask
 * \param Field Pointer to the field object
 * \return The mask
 */
TEMU_API uint64_t temu_getFieldMask(temu_Field *Field);

/*!
 * \brief temu_getFieldValue Retrieve the value of a field in a register
 * \param Obj The object of the class the has the register
 * \param RegName Register name
 * \param FieldName Field name
 * \return The value of the field
 */
TEMU_API uint64_t temu_getFieldValue(void *Obj, const char *RegName,
                                     const char *FieldName);

/*!
 * \brief temu_setFieldValue Set a field's value in a register
 * \param Obj The object that contains the register
 * \param RegName Register name
 * \param FieldName Field name
 * \param Value The value to be set
 * \result zero on success, otherwise non-zero value
 */
TEMU_API int temu_setFieldValue(void *Obj, const char *RegName,
                                const char *FieldName, uint64_t Value);

/*!
 * \brief temu_writeFieldValue Set a field's value in a register with
 * side-effects \param Obj The object that contains the register \param RegName
 * Register name \param FieldName Field name \param Value The value to be set
 * \result zero on success, otherwise non-zero value
 */
TEMU_API int temu_writeFieldValue(void *Obj, const char *RegName,
                                  const char *FieldName, uint64_t Value);

/*!
 * Returns the cold reset value of the register (computed from the field info
 * data). Cold reset is also known as a hard reset, and implies that the power
 * has been off for some time.
 *
 * \param Reg Pointer to the register
 * \return The cold reset value
 */
TEMU_API uint64_t temu_getRegisterColdResetValue(temu_Register *Reg);

/*!
 * Returns the warm reset value of the register
 * \param Reg Pointer to the register
 * \return The warm reset value
 */
TEMU_API uint64_t temu_getRegisterWarmResetValue(temu_Register *Reg);

/*!
 * \brief temu_getRegisterWriteMask Get the current write mask of a register
 * \param Reg Pointer to the register
 * \return The value of the mask
 */
TEMU_API uint64_t temu_getRegisterWriteMask(temu_Register *Reg);

/*!
 * \brief temu_getRegisterReadMask Get the current read mask of a register
 * \param Reg Pointer to the register
 * \return The value of the mask
 */
TEMU_API uint64_t temu_getRegisterReadMask(temu_Register *Reg);

#ifdef __cplusplus
} // extern c
#endif

#endif /* ! TEMU_REGISTER_H */
